import time

def timing(func):
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        time_spent = end_time - start_time
        print("Start Time:", time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(start_time)))
        print("End Time:", time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(end_time)))
        print("Time Spent:", "{:.2f} seconds".format(time_spent))
        return result
    return wrapper


### example usage
if __name__ == "__main__":
    @timing
    def your_function():
        # Your function's code here
        pass

    # Call your decorated function
    your_function()


import os
import time 
import json

from ._SubprocessHandler import run
from ._Timing import timing

# A concise interface function to other module:  
def bulkclone():
    RepoCloner("gpf-config.json")


# Complete structure: 
class RepoCloner:
    def __init__(self, config_file_path):
        self.config_file_path = config_file_path
        self.root_directory = ""
        self.repositories = []
        self.success = []
        self.failed = []
        self.no_effect = []

    @timing
    def listclone(self):
        self.load_config()
        self.clone_repositories()
        self.print_summary()

    def load_config(self):
        with open(self.config_file_path, "r") as config_file:
            config = json.load(config_file)
        self.root_directory = config["root_directory"]
        self.repositories = config["repositories"]

    def clone_repo(self, remote_url, local_path):
        if os.path.exists(local_path) and any(os.listdir(local_path)):
            print(f"Repository '{local_path}' already exists. Skipping cloning.")
            return 2
        command = ["git", "clone", remote_url, local_path]
        run(command, check=True)
        print(f"Repository '{local_path}' cloned successfully.")
        
        git_folder_path = os.path.join(local_path, ".git")
        if not os.path.exists(git_folder_path):
            print(f"Error: '.git' folder was not generated for repository '{local_path}'. Aborting.")
            return 1
        return 0

    def clone_repositories(self):
        for repo_info in self.repositories:
            repo_name = repo_info["name"]
            repo_url = repo_info["remote_url"]
            local_path = os.path.join(self.root_directory, repo_name)
            status_message = self.clone_repo(repo_url, local_path)
            if status_message == 0:
                self.success.append(repo_name)
            elif status_message == 1:
                self.failed.append(repo_name)
            elif status_message == 2:
                self.no_effect.append(repo_name)

    def print_summary(self):
        print("Repositories cloned successfully:", self.success)
        print("Repositories failed to clone:", self.failed)
        print("Repositories already existed:", self.no_effect)

if __name__ == "__main__":
    cloner = RepoCloner("gpf-config.json")
    cloner.listclone()


import os
import subprocess

# A concise interface function to other module:  
def run(cmd: str, loc: str) -> tuple[int, str]:
    return_code, stdout = _SubprocessHandler().run(cmd, loc)
    return return_code, stdout

# Complete structure: 
class _SubprocessHandler:

    def run(self,command, running_location = None):
        """This method runs a command and captures its output and return code."""
        print("=" * 30)
        
        running_location = running_location if running_location is not None else str(os.getcwd()) 
        print(f"Working in the {running_location} folder...") # it was os.path.abspath(self.running_location)
        print(f"Run: {command}")
        
        # there are 3 possible cases: Successful-non-zero, successful-zero, Failed-subprocess. 
        try:
            # This version of subprocess.Popen() could return all message from execution process. 
            result = subprocess.Popen(
                command,
                cwd=running_location,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

            # Capture the standard output and standard error
            stdout, stderr = result.communicate()
            string_stdout = stdout.decode('utf-8')
            string_stderr = stderr.decode('utf-8')
            return_code = result.returncode
            
            # In the current setting, print all outputs from terminal for development purpsoe 
            print("Returned exit code:", return_code)
            if not string_stdout:
                print("Standard Output: <--- It is empty!")
            else:
                print("Standard Output:", string_stdout)
            if not string_stderr:
                print("Standard Error: <--- It is empty!")
            else:
                print("Standard Error:", string_stderr)
            
            # Case 2: Successful-zero
            if return_code == 0:
                print("Command executed successfully!")
            # Case 1: Successful-non-zero
            else:
                print("Command failed. Please check the above messages")
            return return_code, string_stdout
        # Case 3a: If there is any non-zero return code
        except subprocess.CalledProcessError as e:
            print("Subprocess failed with exit code:", e.returncode)
            print("Error output:", e.stderr)
        # Case 3b: Generic catch-all for any other exceptions. 
        except Exception as e:
            print("An error occurred:", e)

# Testing unit: 
if __name__ == "__main__":
    command_to_run = ["ls", "-l"]
    working_location = os.getcwd()
    handler = _SubprocessHandler().run(command_to_run, working_location)
    # return success in 23 Aug 2023


import os
import json

"""
This module is writing a class "ConfigJSONHandler". 
This class's main functionality is the 'check_initial_setup_then_get_datastore_json_address()' method.
Which should be triggered when the user is the first time using the package, or the user just reinstalled the package. 

"""
"""
# Learning journey:

# A good class should store its resources in its attribute, 
# not getting them every time in the running process. 
# That's very good to "keep things into record, as an attribute.
# Especially for the process that could be completed immediately. 
# This class should not depend on other classes / functioos 


# To-do list:
1. Improve error handling - Done. 
2. Spacing: Some of your indentation and spacing inside the prompt strings could be improved for better readability.


"""

# interface functions



class ConfigJSONHandler:

    def __init__(self):
        # all details of config.json should be put at here. 
        self.config_json_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "package_setup.json")
        # return something like: /home/kin/All_Github_Repos/gites/gites/subpackage/package_setup.json
        self.config_data: dict = self._load_config()
        # the second parameter of `.get()` method is the value to be returned if the key is not found in the dictionary. 
        self.json_initial_setup_check = self.config_data.get("initial_setup", False) #
        self.default_datastore_json_name = self.config_data.get("default_datastore_json_name", "my_gites.json")
        self.root_dir_of_gites_datastore_json = self.check_initial_setup_then_get_datastore_json_address()
        self.gites_datastore_json_location = os.path.join(self.root_dir_of_gites_datastore_json, self.default_datastore_json_name)


    def _load_config(self):
        """A private method that loads config json of the package. 
        No inputs needed because the location of the config file is fixed. """
        if os.path.exists(self.config_json_path):
            try:
                with open(self.config_json_path, "r") as json_file:
                    return json.load(json_file) # 
            except json.JSONDecodeError as e:
                print(f"Error decoding JSON: {e}")
                raise SystemExit("Terminating due to unexpected error")
        else:
            print("Config JSON file not found:", self.config_json_path)
            raise RuntimeError(f"Failed to load config data on the location: {self.config_json_path}")

    def _update_config(self, key, value):
        """A private method that update values for a specific key on the config.json.
        The user need must specify which key, which value. 
        """
        config = self.config_data
        if config is not None:
            try:
                config[key] = value
                with open(self.config_json_path, 'w') as json_file:
                    json.dump(config, json_file, indent=4)
                print(f"Updated '{key}' with '{value}' in '{self.config_json_path}'.")
            except Exception as e:
                print(f"An error occurred: {e}")
                raise SystemExit("Terminating due to unexpected error")

    def _get_default_userdata_path(self, json_name):
        """ This private tries to generate a "default location" for our datastore json. 
        "default location" is an option of 'check_initial_setup' when the user is asked whether
        they specify a location, or just use the default location. 
        
        The input variable 'json_name' is pre-defined in config json.
        """
        if os.name == 'nt':  # Windows
            appdata_path = os.environ.get('APPDATA', os.path.expanduser('~'))
            path = os.path.join(appdata_path, 'YourApp')
        elif os.name == 'posix':  # Linux and macOS
            home_path = os.environ.get('HOME', os.path.expanduser('~'))
            path = os.path.join(home_path, '.yourapp')
        else:
            raise RuntimeError("Unsupported operating system")
        # Create the directory if it doesn't exist
        if not os.path.exists(path):
            os.makedirs(path)
        return os.path.join(path, json_name)

    def check_initial_setup_then_get_datastore_json_address(self):
        """This is the main functionality of this class. When gites is initialized. This method should be called. 
        When the user is the first time using this package, or he just updated the package. 
        The user need to specify the location of the datastore json. 
        
        If the user didn't used this package before, this package will create a new datastore json for the user.
        """
        default_datastore_json_name = self.default_datastore_json_name
        default_json_path = self._get_default_userdata_path(default_datastore_json_name)

        prompt = f"""
        You are the first time using the package, or have upgraded the package. 
        If you want to define a customized my_gites.json in the next step, enter y.
        If you want to use the default location, enter n.
        (Default location: {default_json_path})
        
        Please enter (y/n) to proceed.
        """

        # Case 1: If it is already not the first time using the package after (re)installing, 
        # skip all other process and just load the location from config.json. 
        if not self.json_initial_setup_check:
            return self.config_data["gites_datastore_json_location"]
        # Case 2: If it is the first time using the package. Guiding the user to complete the specification. 
        else:
            use_custom_path = input(prompt)
            if use_custom_path.lower() == "y":
                try:
                    custom_path = input(f"Please enter the directory address that contains your {default_datastore_json_name}. "
                                        f"The example is '/home/kin/.yourapp/'. Please do not include the file name. ")
                    self._update_config("gites_datastore_json_location", custom_path)
                    self._update_config("initial_setup", False)
                    return custom_path
                except Exception as e:
                    print(f"Error creating custom path: {e}")
                    raise SystemExit("Terminating due to unexpected error")
            elif use_custom_path.lower() == "n":
                self._update_config("gites_datastore_json_location", default_json_path)
                self._update_config("initial_setup", False)
                return default_json_path
            else:
                print("Invalid input.")
                raise SystemExit("Exiting due to invalid input.")

if __name__ == "__main__":
    datastore_address = ConfigJSONHandler().gites_datastore_json_location


import os
import datetime

from ._SubprocessHandler import run
from .DatastoreJSONHandler import DatastoreJSONHandler

# A concise interface function to other module:  
def bulkpush(list_of_repo):
    return GitPushManager.listpush(list_of_repo)

# Complete structure: 
class GitPushManager:
    def __init__(self):
        # granting all informations: 
        self.datastore_json_handler = DatastoreJSONHandler()
        self.root_folder = self.datastore_json_handler.root_dir
        
        all_items = os.listdir(self.root_folder)
        self.repo_list  = [os.path.join(".", item) for item in all_items if os.path.isdir(os.path.join(self.root_folder, item)) and not item.startswith('.')]
        
        # for statistics and summary: 
        self.success_repo = []
        self.failed_repo = []
        self.no_effect_repo = []
        
        # go to root folder before list pushing. 
        os.chdir(self.root_folder)


    def lpush(self):

        print(f"Totally there are {len(self.repo_list)} repos to work on. They are: {self.repo_list}")
        os.chdir(self.root_folder)

        for repo in self.repo_list:
            print("+" * 72)
            print(f"Current working directory: {os.getcwd()}")
            print(f"Working on repository: {repo}")
            
            # Git checkout
            return_code, _ = run(["git", "checkout", "main"], loc = repo)
            if return_code != 0:
                self.failed_repo.append(repo)
                print("Gheckout failed")
                # Continue: terminate the process for this element, proceed next element in the for-loop
                continue 

            # Git status
            return_code, stdout = run(["git", "status"], loc=repo)
            # 3 cases in 'git status': Case 1: there is something new (no need to stop). Case 2 and 3: stop the `listpush` for that repo.
            # Case 3: Other case 
            if return_code != 0:
                self.failed_repo.append(repo)
                print("Git status command failed.")
                continue # type: ignore
            # Case 2: 'Your branch is up to date'
            if "nothing to commit, working tree clean" in stdout:
                self.no_effect_repo.append(repo)
                print("This repo is new. Proceed the next repo.")
                continue # type: ignore

            # Git add all
            return_code, _ = run(["git", "add", "--all"], loc=repo)
            if return_code != 0:
                self.failed_repo.append(repo)
                print("Failed. Code: a1sd32")
                continue # type: ignore

            # Git commit
            tag_message="Automated add-commit-push"
            timetag_for_commit = datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S")
            commit_command = ["git", "commit", "-m", f"{tag_message}. Datetime tag: {timetag_for_commit}"]
            return_code, stdout = run(commit_command, loc=repo)
            # Case 2: Any other case
            if "Your branch is up to date" in stdout:
                print("No changes to commit.")
                self.no_effect_repo.append(repo)
                print("No effect to the repo. Code: 1sa3a")
                continue
            if return_code != 0:
                self.failed_repo.append(repo)
                print("Failed. Code: 1zsxer2")
                continue # type: ignore
            # Case 3 (Trivial): The staging area is empty (But it is most probably blocked due to the previous `git add` control flow)
            if "nothing to commit" in stdout:
                print("No changes to commit.")
                self.no_effect_repo.append(repo)
                print("No effect to the repo. Code: t5gs2")
                continue # type: ignore
            print(f"Commit successful. Output:\n{stdout}")

            # Git push origin main
            push_command = ["git", "push", "origin", "main"]
            _, push_output = run(push_command, loc=repo)
            if "Total" in push_output or not push_output:
                self.success_repo.append(repo)
                # Future development: Scrape the terminal output and collect those data to summary variable.
                print("Push completed")
            else:
                print(f"Push failed. Output:\n{push_output}")
                self.failed_repo.append(repo)
            print("=" * 72)

            
        print("|" * 72)
        print("Add-commit-push completed for all repos in the root folder. Here is the work summary: ")
        print(f"Successful repos (Totally {len(self.success_repo)}): {self.success_repo}")
        print(f"Failed repos (Totally {len(self.failed_repo)}): {self.failed_repo}")
        print(f"No effect repos (Totally {len(self.no_effect_repo)}): {self.no_effect_repo}")

# Testing unit: 
if __name__ == "__main__":
    GitPushManager().lpush()




"""
This is a private module that encapsulate the communication process from the JSONUpdater. 
This module should not contains main functions that the package would use. 
"""

import os
import json
from .ConfigJSONHandler import ConfigJSONHandler



### Complete structure: 
class DatastoreJSONHandler:
    def __init__(self):
        self.config_json_handler = ConfigJSONHandler()
        # the following are the data from the datastore json: 
        self.gites_datastore_json_location = self.config_json_handler.gites_datastore_json_location # this is for naming convention.
        self.data = self.load_datastore_json()
        self.list_of_repo_details = self.data["repositories"] # this return a list of dictionary, containing 'name' and 'remote_url' key for each repo.
        self.root_dir = self.data["root_directory"]

    def load_datastore_json(self):
        existing_data = {}

        # case 1: that location has exactly-matched file
        if os.path.exists(self.gites_datastore_json_location):
            with open(self.gites_datastore_json_location, "r") as json_file:
                try:
                    existing_data = json.load(json_file)
                except json.JSONDecodeError as e:
                    print(f"Error loading JSON: {e}")
        # case 2a: after opening the file, the outermost structure is not a json
        if not isinstance(existing_data, dict):
            print("Please check the file. The whole file is not detected as a json file.")
            raise SystemExit("Terminating due to unexpected error")
        # case 2b: Check each item in the "repositories" data to verify it is well-structured. 
        for repo_detail in existing_data["repositories"]:
            if not isinstance(repo_detail, dict):
                print("Please check the file. The repository list of the file is not detected as a json file.")
                raise SystemExit("Terminating due to unexpected error")
        # case 3: after knowing it is a json, but the json has empty repo entries
        if "repositories" not in existing_data:
            print("it is an empty json. creating structure for it...")
            existing_data["repositories"] = []
        # finally return data
        return existing_data

# Testing unit: 
if __name__ == "__main__":
    # this module need an existing json, or existing data, to perform action. 
    # default_json_path = DatastoreLocationChecker().load_setup_json_and_get_datastore_json_address()
    # json_existing_data = _JSONHandler(default_json_path)
    # root_dir = json_existing_data.data.get("root_directory")
    # this is just a default location. 
    pass


"""
This cli.py is just an interface that collect and manage all the actions,
which help centralize the logic for handling variaous commands and make it easier to maintain 
and expand in the future. 

There are two parts need to be edited to make CLI works. 
1. cli.py: it is an adapter between CLI listener in `setup.py` and the functionalities from subpackages.
argparse module parse 'subcommands' from command-line arguments. 


2. setup.py: the `entry_points` attribute setup the listeners and the related keywords.
'gites' considered as entry-point and the first keyword of the command to invoke the package.
"""


from .subpackage.ConfigJSONHandler import ConfigJSONHandler
from .subpackage.GitPushManager import GitPushManager

import argparse

def _check_datastore_location():
    """This function would be optionally used by multiple CLI actions. """
    datastore_json_path = ConfigJSONHandler.check_initial_setup_then_get_datastore_json_address()

def cli_lpush():
    GitPushManager().lpush() 

def main():
    parser = argparse.ArgumentParser(description='Command-line interface for gites package')
    # subparsers handle subcommands. 
    subparsers = parser.add_subparsers(dest='command', help='Available commands')

    # Create a subparser for the 'push' command
    push_parser = subparsers.add_parser('lpush', help='Push changes to Git')
    push_parser.set_defaults(func=cli_lpush)

    args = parser.parse_args()

    if hasattr(args, 'func'):
        args.func()

"""
Also check the CLI entry point script at <package_root_folder>/bin/gites.
There should be a shebang in the first line, to be something like: #!/usr/bin/env python
"""

# The content of that file would be like:
"""
#!/home/kin/anaconda3/bin/python
# -*- coding: utf-8 -*-
import re
import sys
from gites.cli import cli_lpush
if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw|\.exe)?$', '', sys.argv[0])
    sys.exit(cli_lpush())
"""


if __name__ == "__main__":
    cli_lpush()





import os
import util
from ._SubprocessHandler import _SubprocessHandler
from .DatastoreJSONHandler import DatastoreJSONHandler


# Complete structure: 
class CreateRepoManager:
    def __init__(self, new_created_repo_folder_name, new_created_repo_url):
        self.datastore_json_handler = DatastoreJSONHandler()
        
        # new repo information: 
        self.new_created_repo_folder_name = new_created_repo_folder_name
        self.new_created_repo_url = new_created_repo_url

    def run(self, cmd):
        _SubprocessHandler().run(cmd)

    def check_remote_origin(self):
        result = self.run(['git', 'remote', '-v'])
        if result is not None:
            _, result_stdout = result
            return repo_url in result_stdout
        return False

    def initialize_repo(self):
        
        # Synthesize the path for the newly created repo.
        datastore_json_data = self.datastore_json_handler.data
        root_folder = datastore_json_data["root_directory"]
        full_folder_location = os.path.join(root_folder, self.new_created_repo_folder_name)

        # Go to that folder. If it is not existed, create it. 
        if not os.path.exists(full_folder_location):
            os.makedirs(full_folder_location)
        os.chdir(full_folder_location)
        
        git_folder = os.path.join(full_folder_location, ".git")
        if os.path.exists(git_folder):
            print(f"Deleting existing .git folder at {git_folder}")
            cmd = ['rm', '-rf', git_folder]
            self.run(cmd)

        commands_set_1 = [
            ['git', 'init'],
        ]
        commands_set_2 = [
            ['git', 'add', '.'],
            ['git', 'commit', '-m', 'first commit'],
            ['git', 'checkout', '-b', 'main'],
        ]
        commands_set_3 = [
            ['git', 'remote', '-v'],
            ['git', 'push', '-u', 'origin', 'main'],
        ]

        # create readme.md
        file_path = 'readme.md'
        if not os.path.exists(file_path):
            with open(file_path, 'w') as f:
                f.write('initial document\n')
        else:
            print("File 'readme.md' already exists, skipping creation.")

        for cmd in commands_set_1:
            self.run(cmd)

        util.check_and_copy_pre_commit_hook(full_folder_location)

        for cmd in commands_set_2:
            self.run(cmd)

        list_of_remote_url = self.check_remote_origin()

        if list_of_remote_url:
            self.run(['git', 'remote', 'set-url', 'origin', list_of_remote_url])
        else:
            self.run(['git', 'remote', 'add', 'origin', repo_url])


        for cmd in commands_set_3:
            self.run(cmd)

if __name__ == "__main__":
    # initialize it: 

    # future development: make it as input() field to let the user just copy things here. 
    repo_folder_name = "Factual Value Asset" # Synthesis the full path with the "root_directory" value in the datastore json.
    repo_url = "https://github.com/pakkinlau/Factual-Value-Asset.git"

    create_repo_manager = CreateRepoManager(repo_folder_name, repo_url).initialize_repo()